# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of Yaji.
#
# Yaji is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yaji is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Yaji.  If not, see <http://www.gnu.org/licenses/>.

from typing import *
import threading

from .core import log


class AppConfig:
    """
    Application configuration storage.

    This can store arbitrary configuration items, which then become available on browser side as well.
    """

    def __init__(self, app: 'Application'):
        self.__cfg = {}
        self.__app = app
        self.__appinitialized = False
        self.__lock = threading.Lock()

    def __markinitialized(self) -> None:
        """
        Marks the configuration as initialized (in order to handle requests a bit differently).
        |infrastructure|
        """
        self.__appinitialized = True
        log.debug("AppConfig marked as initialized keeping %s.", self.__cfg)

    def getconfigs(self, _markinitialized: bool = False) -> Dict[str, Any]:
        """
        Returns the complete configuration.
        |infrastructure|

        :param _markinitialized: If to consider the configuration as initialized by this call.
        """
        with self.__lock:
            if _markinitialized:
                self.__markinitialized()
            return dict(self.__cfg)

    def getconfig(self, key: str, default: Optional[Any] = None) -> Any:
        """
        Returns a configuration value by key.

        :param key: The configuration key.
        :param default: The default value (if no such key).
        """
        with self.__lock:
            return self.__cfg.get(key, default)

    def setconfig(self, key: str, value: Any) -> None:
        """
        Sets a configuration value by key.

        :param key: The configuration key.
        :param value: The new value.
        """
        with self.__lock:
            self.__cfg[key] = value
            if self.__appinitialized:
                self.__app.triggerevent("_yj_appconfig_set", key=key, value=value)


if TYPE_CHECKING:  # ugly hack for type annotations
    from .app import *
