# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of Yaji.
#
# Yaji is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yaji is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Yaji.  If not, see <http://www.gnu.org/licenses/>.

from typing import *
import threading
import time

from .core import log


class BrowserHook:
    """
    Helper for opening browsers.
    |infrastructure|
    """

    def __init__(self, url: str, app: 'Application'):
        """
        :param url: The url to open.
        """
        self.__url = url
        self.__lastheartbeat = None
        self.__app = app
        self.__stop = False
        self.__triedforheartbeatx = 0
        self.__heartbeatx = None

    def heartbeat(self) -> None:
        """
        Gives a browser heartbeat (so the application code side knows that the browser is still alive).
        """
        self.__lastheartbeat = time.time()
        log.debug("BrowserHook received heartbeat.")

    def start(self) -> None:
        """
        Starts the hook.
        """
        # noinspection PyProtectedMember
        def _reopenerthread():
            self.__stop = False
            while not self.__stop:
                if self.__app.takenover:
                    while self.__app.takenover:
                        time.sleep(1)
                    time.sleep(30)
                heartbeatthreshold = self.__app.browser_hook_heartbeat_threshold
                if (self.__lastheartbeat is None) or (time.time() - self.__lastheartbeat) > heartbeatthreshold:
                    if self.__heartbeatx != self.__lastheartbeat:
                        self.__heartbeatx = self.__lastheartbeat
                        self.__triedforheartbeatx = 1
                    elif self.__triedforheartbeatx > 1:
                        self.__app.onopenbrowsererror()
                        break
                    else:
                        self.__triedforheartbeatx += 1
                    if self.__lastheartbeat:
                        if self.__app.stop_implicitly_when_browser_closed:
                            self.__app.stop()
                            break
                        self.__app._browser_was_reopened()
                    self.__app._openbrowser(self.__url)
                    time.sleep(30)
                time.sleep(15)
            log.debug("BrowserHook stopped.")
        threading.Thread(target=_reopenerthread, daemon=True).start()
        log.debug("BrowserHook started.")

    def stop(self) -> None:
        """
        Stops the hook.
        """
        self.__stop = True


if TYPE_CHECKING:  # ugly hack for type annotations
    from .app import *
