# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of Yaji.
#
# Yaji is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yaji is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Yaji.  If not, see <http://www.gnu.org/licenses/>.

from typing import *
import sys
import threading
import weakref


class DataStore:
    """
    A clove-like datastore which then becomes available on browser side as well.
    """

    _lock = threading.Lock()
    _nextid = 0
    _nodedict = {}

    def __init__(self, name: str, app: 'Application', *, curtain: Union[bool, Optional[str]] = None, _editor_for=None):
        """
        |infrastructure|
        """
        self.__name = name
        self.__app = app
        self.__curtain = curtain
        self.__editor_for = _editor_for
        self.__rootnode = DataStore.Node(self)
        self.__lock = threading.RLock()
        self.__onchanged_handlers = []

    @staticmethod
    def getnodebyid(ajaxid: int) -> Optional['DataStore.Node']:
        """
        Returns the node for an ajax id.
        """
        with DataStore._lock:
            if ajaxid not in DataStore._nodedict:
                return None
            x = DataStore._nodedict[ajaxid]()
            if x is None:
                DataStore._nodedict.pop(ajaxid)
            return x

    class ValuePointer:
        """
        A value pointer is kind of an address or reference to a node (i.e. a location) inside a datastore.

        .. note::
           The root node is usually referenced by :samp:`None`.
        """

        def __init__(self, irow: int, icol: int, node: 'DataStore.Node'):
            """
            |infrastructure|
            """
            self.irow = irow
            self.icol = icol
            self.node = node

    class Node:
        """
        A location inside a datastore, having a value and potentially a 2-dimensional grid of child nodes.
        """

        def __init__(self, datastore: 'DataStore'):
            """
            |infrastructure|
            """
            with DataStore._lock:
                self.ajaxid = DataStore._nextid
                DataStore._nextid += 1
                DataStore._nodedict[self.ajaxid] = weakref.ref(self)
            self.datastore = datastore
            self.value = None
            self.irow = 0
            self.icol = 0
            self.rowcount = 0
            self.colcount = 0
            self.childs = []
            self.parent = None

        def appendrow(self, values: Optional[List[Optional[Any]]] = None) -> Optional['DataStore.Node']:
            """
            Appends a row (like inserting to the end).

            :param values: An optional list of values to put into the new nodes.
            """
            with self.datastore:
                return self.datastore.appendrow(self.valuepointer, values)

        def appendcolumn(self, values: Optional[List[Optional[Any]]] = None) -> Optional['DataStore.Node']:
            """
            Appends a column (like inserting to the end).

            :param values: An optional list of values to put into the new nodes.
            """
            with self.datastore:
                return self.datastore.appendcolumn(self.valuepointer, values)

        def insertrow(self, i: int, values: Optional[List[Optional[Any]]] = None) -> Optional['DataStore.Node']:
            """
            Inserts a row.

            :param i: The position to insert the row to.
            :param values: An optional list of values to put into the new nodes.
            """
            with self.datastore:
                return self.datastore.insertrow(i, self.valuepointer, values)

        def insertcolumn(self, i: int, values: Optional[List[Optional[Any]]] = None) -> Optional['DataStore.Node']:
            """
            Inserts a column.

            :param i: The position to insert the column to.
            :param values: An optional list of values to put into the new nodes.
            """
            with self.datastore:
                return self.datastore.insertcolumn(i, self.valuepointer, values)

        def removerow(self, i: int) -> None:
            """
            Removes a row.

            :param i: The position to remove the row from.
            """
            with self.datastore:
                self.datastore.removerow(i, self.valuepointer)

        def removecolumn(self, i: int) -> None:
            """
            Removes a column.

            :param i: The position to remove the column from.
            """
            with self.datastore:
                self.datastore.removecolumn(i, self.valuepointer)

        def getvalue(self) -> Optional[Any]:
            """
            Returns the value of this node.
            """
            with self.datastore:
                return self.datastore.getvalue(self.valuepointer)

        def setvalue(self, value: Optional[Any] = None) -> None:
            """
            Sets the value of this node.

            :param value: The new value.
            """
            with self.datastore:
                self.datastore.setvalue(value, ptr=self.valuepointer)

        @property
        def valuepointer(self) -> 'DataStore.ValuePointer':
            """
            Returns a value pointer for this node.
            """
            with self.datastore:
                if self.parent:
                    return self.datastore.valuepointer(self.irow, self.icol, self.parent.valuepointer)

        def _to_simple_repr_(self):
            return '__YajiDataStore_Node', {'value': self.value, 'childs': self.childs,
                                            'rowcount': self.rowcount, 'colcount': self.colcount}

    def __trigger_update_event(self, node: 'DataStore.Node') -> None:
        """
        Triggers an update event.

        :param node: The node that was updated.
        """
        for handler in self.__onchanged_handlers:
            handler(node)
        if self.__app:
            self.__app.triggerevent("_yj_datastore_update", datasource=self.name, id=self.node_to_idpath(node))

    def add_onchanged_handler(self, handler: Callable[['DataStore.Node'], None]) -> None:
        """
        Add a custom handler for changes in this datastore.

        :param handler: A handler function.
        """
        self.__onchanged_handlers.append(handler)

    @property
    def name(self) -> str:
        """
        The datastore name.
        """
        return self.__name

    @property
    def curtain(self) -> Union[bool, Optional[str]]:
        """
        The datastore curtain expression (pointing to a yaji.Curtain in browser side javascript).
        """
        return self.__curtain

    @property
    def rootnode(self) -> 'DataStore.Node':
        """
        The root node.
        """
        return self.__rootnode

    def rowcount(self, ptr: Optional['DataStore.ValuePointer'] = None) -> int:
        """
        Returns the row count of a node (by ValuePointer).

        :param ptr: The value pointer of the node to query.
        """
        return self.ptr_to_node(ptr).rowcount

    def columncount(self, ptr: Optional['DataStore.ValuePointer'] = None) -> int:
        """
        Returns the column count of a node (by ValuePointer).

        :param ptr: The value pointer of the node to query.
        """
        return self.ptr_to_node(ptr).colcount

    def appendrow(self, ptr: Optional['DataStore.ValuePointer'] = None,
                  values: Optional[List[Optional[Any]]] = None) -> Optional['Node']:
        """
        Appends a row (like inserting to the end).

        :param ptr: The parent node (by ValuePointer).
        :param values: An optional list of values to put into the new nodes.
        """
        return self.insertrow(sys.maxsize, ptr, values)

    def appendcolumn(self, ptr: Optional['DataStore.ValuePointer'] = None,
                     values: Optional[List[Optional[Any]]] = None) -> Optional['Node']:
        """
        Appends a column (like inserting to the end).

        :param ptr: The parent node (by ValuePointer).
        :param values: An optional list of values to put into the new nodes.
        """
        return self.insertcolumn(sys.maxsize, ptr, values)

    def insertrow(self, i: int, ptr: Optional['DataStore.ValuePointer'] = None,
                  values: Optional[List[Optional[Any]]] = None) -> Optional['Node']:
        """
        Inserts a row.

        :param i: The position to insert the row to.
        :param ptr: The parent node (by ValuePointer).
        :param values: An optional list of values to put into the new nodes.
        """
        with self.__lock:
            if values is None:
                values = []
            pnode = self.ptr_to_node(ptr)
            i = min(i, pnode.rowcount)
            pnode.childs.insert(i, [])
            for j in range(i + 1, pnode.rowcount):
                for jc in pnode.childs[j]:
                    jc.irow = j
            pnode.rowcount += 1
            for jv in range(len(values)):
                if self.columncount(ptr) <= jv:
                    self.insertcolumn(jv, ptr)
                self.setvalue(values[jv], ptr=self.valuepointer(i, jv, ptr))
            self.__trigger_update_event(pnode)
            rr = pnode.childs[i] if (len(pnode.childs) > i) else None
            return rr[0] if (rr and len(rr) > 0) else None

    def insertcolumn(self, i: int, ptr: Optional['DataStore.ValuePointer'] = None,
                     values: Optional[List[Optional[Any]]] = None) -> Optional['Node']:
        """
        Inserts a column.

        :param i: The position to insert the column to.
        :param ptr: The parent node (by ValuePointer).
        :param values: An optional list of values to put into the new nodes.
        """
        with self.__lock:
            if values is None:
                values = []
            pnode = self.ptr_to_node(ptr)
            i = min(i, pnode.colcount)
            for jr in pnode.childs:
                if i < len(jr):
                    jr.insert(i, None)
                    for j in range(i + 1, len(jr)):
                        jr[j].icol = j
            pnode.colcount += 1
            for jv in range(len(values)):
                if self.rowcount(ptr) <= jv:
                    self.insertrow(jv, ptr)
                self.setvalue(values[jv], ptr=self.valuepointer(jv, i, ptr))
            self.__trigger_update_event(pnode)
            rr = pnode.childs[0] if (len(pnode.childs) > 0) else None
            return rr[i] if (rr and len(rr) > i) else None

    def removerow(self, i: int, ptr: Optional['DataStore.ValuePointer'] = None) -> None:
        """
        Removes a row.

        :param i: The position to remove the row from.
        :param ptr: The parent node (by ValuePointer).
        """
        with self.__lock:
            pnode = self.ptr_to_node(ptr)
            pnode.childs.pop(i)
            for j in range(i, pnode.rowcount):
                for jc in pnode.childs[j]:
                    jc.irow = j
            pnode.rowcount -= 1
            self.__trigger_update_event(pnode)

    def removecolumn(self, i: int, ptr: Optional['DataStore.ValuePointer'] = None) -> None:
        """
        Removes a column.

        :param i: The position to remove the column from.
        :param ptr: The parent node (by ValuePointer).
        """
        with self.__lock:
            pnode = self.ptr_to_node(ptr)
            for jr in pnode.childs:
                if i < len(jr):
                    jr.pop(i)
                    for j in range(i, len(jr)):
                        jr[j].icol = j
            pnode.colcount -= 1
            self.__trigger_update_event(pnode)

    def getvalue(self, ptr: Optional['DataStore.ValuePointer'] = None) -> Optional[Any]:
        """
        Returns the value of a node (by ValuePointer).

        :param ptr: The value pointer of the node to query.
        """
        return self.ptr_to_node(ptr).value

    def node_to_idpath(self, node: 'DataStore.Node') -> List[int]:
        """
        Returns a path of node ids from a given node up the hierarchy to root (as list).

        :param node: The node to query.
        """
        with self.__lock:
            result = []
            while node != self.rootnode:
                result.append(node.ajaxid)
                node = node.parent
            return result

    def setvalue(self, value: Optional[Any] = None, *, ptr: Optional['DataStore.ValuePointer'] = None) -> None:
        """
        Sets the value of a node (by ValuePointer).

        :param value: The new value.
        :param ptr: The value pointer of the node to modify.
        """
        with self.__lock:
            pnode = self.ptr_to_node(ptr)
            pnode.value = value
            self.__trigger_update_event(pnode)

    def parent(self, ptr: Optional['DataStore.ValuePointer'] = None) -> 'DataStore.ValuePointer':
        """
        Returns the parent ValuePointer for a given ValuePointer.

        :param ptr: The value pointer of the node to query.
        """
        with self.__lock:
            pnode = self.ptr_to_node(ptr)
            return DataStore.ValuePointer(pnode.irow, pnode.icol, pnode)

    def ptr_to_node(self, ptr: Optional['DataStore.ValuePointer']) -> 'DataStore.Node':
        """
        Returns the Node for a ValuePointer.

        :param ptr: The value pointer of the node to query.
        """
        return ptr.node if ptr else self.rootnode

    def node_to_ptr(self, node: 'DataStore.Node') -> Optional['DataStore.ValuePointer']:
        """
        Returns the ValuePointer for a Node.

        :param node: The node to query.
        """
        with self.__lock:
            result = None
            cpath = []
            cnode = node
            while cnode != self.rootnode:
                cpath.append(cnode)
                cnode = cnode.parent
            for dnode in reversed(cpath):
                result = self.valuepointer(dnode.irow, dnode.icol, result)
            return result

    def valuepointer(self, irow: int, icol: int,
                     parent: Optional['DataStore.ValuePointer']) -> 'DataStore.ValuePointer':
        """
        Returns a ValuePointer by a parent ValuePointer and a row and a column index.

        :param irow: The row index.
        :param icol: The column index.
        :param parent: The value pointer to the parent node.
        """
        with self.__lock:
            pnode = self.ptr_to_node(parent)
            if irow >= pnode.rowcount or icol >= pnode.colcount:
                raise Exception("bad pointer")
            row = pnode.childs[irow]
            while len(row) <= icol:
                row.append(None)
            if row[icol] is None:
                row[icol] = DataStore.Node(self)
            node = row[icol]
            node.parent = pnode
            node.irow = irow
            node.icol = icol
            return DataStore.ValuePointer(irow, icol, node)

    def _to_simple_repr_(self):
        if self.__editor_for:
            # noinspection PyProtectedMember
            return self.__editor_for._to_simple_repr_()
        else:
            return '__Datasource', {'islocal': False, 'name': self.name, 'curtain': self.curtain}

    def __enter__(self):
        self.__lock.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__lock.release()


if TYPE_CHECKING:  # ugly hack for type annotations
    from .app import *
