# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of Yaji.
#
# Yaji is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yaji is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Yaji.  If not, see <http://www.gnu.org/licenses/>.

from typing import *

from .datastore import DataStore


class BrowserSideDatasource:
    """
    A browser side datasource. You can only edit it before pushing it to browser side, and only inside a :samp:`with`
    block.
    """

    def __init__(self, name: str):
        self.__name = name
        self.__init_with = None
        self.__datastore = None

    @property
    def name(self) -> str:
        """
        The data source name.
        """
        return self.__name

    def _to_simple_repr_(self):
        return '__Datasource', {'islocal': True, 'name': self.name, 'init_with': self.__init_with}

    def __enter__(self):
        # noinspection PyTypeChecker
        self.__datastore = DataStore(self.__name, None, _editor_for=self)
        return self.__datastore

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__init_with = self.__datastore.rootnode
        self.__datastore = None


class BindDirection:
    """
    Data flow directions of data bindings.
    """

    #:  Bidirectional data transfer.
    Bidirectional = "bidirectional"

    #:  Data transfer only from datasource to widget.
    ToWidget = "towidget"

    #:  Data transfer only from widget to datasource.
    ToDatasource = "todatasource"


class Visibility:
    """
    The visibility of a widget.
    """

    #:  Visible.
    Visible = "clove_visible"

    #:  Invisible but taking space.
    Invisible = "clove_invisible"

    #:  Invisible and collapsed.
    InvisibleCollapsed = "clove_invisiblecollapsed"


class AbstractFunction:
    """
    An abstract callable function.
    """
    pass


class BackendFunction(AbstractFunction):
    """
    A callable application code side function.
    """

    def __init__(self, url: str, funcargs: Optional[Dict[str, Any]] = None):
        """
        :param url: The url to trigger for executing this action.
        :param funcargs: Function arguments.
        """
        super().__init__()
        self.__url = url
        self.__funcargs = funcargs or {}

    def _to_simple_repr_(self):
        return '__BackendFunction', {'url': self.__url, 'funcargs': self.__funcargs}


class BrowserFunction(AbstractFunction):
    """
    A callable browser side function.
    """

    def __init__(self, funcname: str, funcargs: Optional[List[Any]] = None):
        """
        :param funcname: The name of the function to trigger on browser side.
        :param funcargs: Function arguments.
        """
        super().__init__()
        self.__funcname = funcname
        self.__funcargs = funcargs or []

    def _to_simple_repr_(self):
        return '__BrowserFunction', {'funcname': self.__funcname, 'funcargs': self.__funcargs}


class AbstractAction:
    """
    Abstract menu action (can be an actual action, submenus, separators, ...).
    """

    def __init__(self, *, label: str, icon: Optional['Icon'], checkable: bool, checked: bool, disabled: bool,
                 invisible: bool):
        """
        :param label: The label of this action.
        :param icon: The action icon.
        :param checkable: If this action is checkable (i.e. has a checkbox in menu).
        :param checked: If this action is checked.
        :param disabled: If this action is disabled (cannot be triggered).
        :param invisible: If this action is invisible.
        """
        self.__label = label
        self.__icon = icon
        self.__checkable = checkable
        self.__checked = checked
        self.__disabled = disabled
        self.__invisible = invisible

    def _to_simple_repr_(self):
        return None, {'label': self.__label, 'checkable': self.__checkable, 'checked': self.__checked,
                      'disabled': self.__disabled, 'icon': self.__icon, 'invisible': self.__invisible}


class Action(AbstractAction):
    """
    A menu action.
    """

    def __init__(self, label: str, function: AbstractFunction, *, icon: Optional['Icon'] = None,
                 checkable: bool = False, checked: bool = False, disabled: bool = False, invisible: bool = False):
        """
        :param label: The label of this action.
        :param function: The function to call for this action.
        :param icon: The action icon.
        :param checkable: If this action is checkable (i.e. has a checkbox in menu).
        :param checked: If this action is checked.
        :param disabled: If this action is disabled (cannot be triggered).
        :param invisible: If this action is invisible.
        """
        super().__init__(label=label, icon=icon, checkable=checkable, checked=checked, disabled=disabled,
                         invisible=invisible)
        self.__function = function

    def _to_simple_repr_(self):
        return '__Action', {'function': self.__function, **super()._to_simple_repr_()[1]}


class Submenu(AbstractAction):
    """
    A submenu of a menu.
    """

    def __init__(self, label: str, *, icon: Optional['Icon'] = None, disabled: bool = False, invisible: bool = False):
        """
        :param label: The label of this action.
        :param icon: The action icon.
        :param disabled: If this action is disabled (cannot be triggered).
        :param invisible: If this action is invisible.
        """
        super().__init__(label=label, icon=icon, checkable=False, checked=False, disabled=disabled, invisible=invisible)

    def _to_simple_repr_(self):
        return '__Submenu', super()._to_simple_repr_()[1]


class Separator(AbstractAction):
    """
    A menu separator.
    """

    def __init__(self):
        super().__init__(label="", icon=None, checkable=False, checked=False, disabled=False, invisible=False)

    def _to_simple_repr_(self):
        return '__Separator', super()._to_simple_repr_()[1]


class Icon:
    """
    Icon.
    """

    def __init__(self, by_url: Optional[str] = None, by_symbol: Optional[str] = None):
        """
        :param by_url: A icon by image source url.
        :param by_symbol: A icon by symbol character.
        """
        if by_url is not None:
            self.__src = by_url
            self.__srcfunc = 'byUrl'
        elif by_symbol is not None:
            self.__src = by_symbol
            self.__srcfunc = 'bySymbol'
        else:
            self.__src = None
            self.__srcfunc = None

    @property
    def src(self) -> str:
        """
        The icon source.
        """
        return self.__src

    @property
    def srcfunc(self) -> str:
        """
        The icon source function.
        """
        return self.__srcfunc

    def _to_simple_repr_(self):
        return '__Icon', {'srcfunc': self.srcfunc, 'src': self.src}
