# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of Yaji.
#
# Yaji is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yaji is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Yaji.  If not, see <http://www.gnu.org/licenses/>.

from typing import *
import os
import sys


class PyHtmlView:
    """
    This optional functionality allows to present the user interface of the application in an own bare window instead
    of the usual browser. It is used by default if not configured otherwise and if the system is compatible to it.

    It requires PyQt to be available.
    """

    @staticmethod
    def openview(*, show_url: str, icon: str, wndclassname: str, scriptedclosing: Optional[str],
                 window_width: Optional[int] = None, window_height: Optional[int] = None,
                 window_maximized: bool = False):
        """
        Opens a new view with some content.

        :param show_url: The url to show content from.
        :param icon: The window icon.
        :param wndclassname: The window class name.
        :param scriptedclosing: A javascript expression to execute for closing.
        :param window_width: Width of the window in pixels.
        :param window_height: Height of the window in pixels.
        :param window_maximized: If to maximize the window.
        """
        os.environ['QTWEBENGINE_DISABLE_SANDBOX'] = '1'  # allows usage as root
        import PyQt5.QtCore as QtCore
        import PyQt5.QtGui as QtGui
        import PyQt5.QtWidgets as QtWidgets
        import PyQt5.QtWebEngineWidgets as QtWebEngineWidgets   # if changed, also change in is_system_compatible()!

        class CredentialsDialog(QtWidgets.QDialog):

            def __init__(self, qs, parent):
                super().__init__(parent)
                self.setWindowTitle("Credentials needed")
                vlay = QtWidgets.QFormLayout(self)
                vlay.setFieldGrowthPolicy(vlay.ExpandingFieldsGrow)
                lbl = QtWidgets.QLabel(qs)
                lbl.setWordWrap(True)
                vlay.addRow(lbl)
                self.edtuser = QtWidgets.QLineEdit(self)
                vlay.addRow("User name:", self.edtuser)
                self.edtpasswd = QtWidgets.QLineEdit(self)
                self.edtpasswd.setEchoMode(QtWidgets.QLineEdit.Password)
                vlay.addRow("Password:", self.edtpasswd)
                btns = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok)
                # noinspection PyUnresolvedReferences
                btns.accepted.connect(self.accept)
                vlay.addRow(btns)

            @classmethod
            def ask(cls, qs, parent):
                dlg = CredentialsDialog(qs, parent)
                dlg.exec()
                return dlg.edtuser.text(), dlg.edtpasswd.text()

        class WebPage(QtWebEngineWidgets.QWebEnginePage):

            def __init__(self, profile, parent):
                super().__init__(profile, parent)
                self._parent = parent
                # noinspection PyUnresolvedReferences
                self.authenticationRequired.connect(self.authentication_required)
                # noinspection PyUnresolvedReferences
                self.proxyAuthenticationRequired.connect(self.proxy_authentication_required)

            def _http_auth(self, qs, authenticator):
                username, password = CredentialsDialog.ask(f"Please enter credentials for {qs}", self._parent)
                authenticator.setUser(username)
                authenticator.setPassword(password)

            def authentication_required(self, qurl, authenticator):
                srealm = f" ({authenticator.realm()})" if authenticator.realm() else ""
                self._http_auth(f"{qurl.toString()}{srealm}", authenticator)

            def proxy_authentication_required(self, qurl, authenticator, proxy_host):
                self._http_auth(f"{qurl.toString()} for proxy {proxy_host}", authenticator)

        class WebView(QtWebEngineWidgets.QWebEngineView):

            def __init__(self, *a, **b):
                super().__init__(*a, **b)
                self.setPage(WebPage(None, self))

            def contextMenuEvent(self, event):
                allowedactions = [self.page().action(QtWebEngineWidgets.QWebEnginePage.Cut),
                                  self.page().action(QtWebEngineWidgets.QWebEnginePage.Copy),
                                  self.page().action(QtWebEngineWidgets.QWebEnginePage.Paste),
                                  self.page().action(QtWebEngineWidgets.QWebEnginePage.Undo),
                                  self.page().action(QtWebEngineWidgets.QWebEnginePage.Redo),
                                  self.page().action(QtWebEngineWidgets.QWebEnginePage.PasteAndMatchStyle)]
                menu = self.page().createStandardContextMenu()
                for act in menu.actions():
                    if act not in allowedactions:
                        menu.removeAction(act)
                menu.exec(event.globalPos())

        class MainWindow(QtWidgets.QMainWindow):

            def __init__(self):
                QtWidgets.QMainWindow.__init__(self)
                self.forceclose = False
                self.setMinimumSize(QtCore.QSize(640, 480))
                if window_maximized:
                    self.showMaximized()
                else:
                    self.resize(window_width or 1200, window_height or 660)
                self.setWindowTitle(" ")
                self.setWindowIcon(QtGui.QIcon(icon))
                centralwidget = QtWidgets.QWidget(self)
                self.setCentralWidget(centralwidget)
                gridlayout = QtWidgets.QGridLayout(centralwidget)
                gridlayout.setContentsMargins(0, 0, 0, 0)
                centralwidget.setLayout(gridlayout)
                self.webview = webview = WebView(self)
                # noinspection PyUnresolvedReferences
                webview.titleChanged.connect(lambda t: self.setWindowTitle(t))
                v = 3  # at early times, the 'window' object isnt fine
                def _loadstarted():
                    nonlocal v
                    if v > 0:
                        v -= 1
                        self.webview.page().runJavaScript("window.pyhtmlview=true")
                # noinspection PyUnresolvedReferences
                webview.loadProgress.connect(_loadstarted)
                gridlayout.addWidget(webview, 0, 0)
                webview.load(QtCore.QUrl(show_url))

            def closeEvent(self, event):
                if scriptedclosing and not self.forceclose:
                    event.ignore()
                    def oncloseresult(res):
                        if res:
                            self.forceclose = True
                            self.close()
                    self.webview.page().runJavaScript(scriptedclosing, oncloseresult)
                else:
                    event.accept()

        app = QtWidgets.QApplication(sys.argv)
        app.setApplicationName(wndclassname)
        mainwnd = MainWindow()
        mainwnd.show()
        app.exec_()
        return 0

    @staticmethod
    def is_system_compatible() -> bool:
        """
        Tests if your system is compatible with PyHtmlView.
        """
        try:
            import PyQt5.QtWebEngineWidgets
            return True
        except ImportError:
            return False
