# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of Yaji.
#
# Yaji is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yaji is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Yaji.  If not, see <http://www.gnu.org/licenses/>.

from typing import *
import json
import threading

from .i18n import TrStr, TrStrOrStrTyping
from .request import Request


class UserFeedbackController:
    """
    A controller for user feedback operations.

    Use methods inside it for communicating with the user during some operations.
    Such an object is automatically available during execution, there is no need to create new ones.
    """

    def __init__(self, app: 'Application'):
        """
        |infrastructure|
        """
        self.__nextreqid = 0
        self.__responses = {}
        self.__app = app
        self.__responseslock = threading.Lock()
        self.__responsescond = threading.Condition(self.__responseslock)
        self.__nix = object()

    def make_raw_request(self, kind: str, **params: Any) -> Any:
        """
        Makes a low-level userfeedback request.

        :param kind: The raw userfeedback kind name as known on browser side. You can register custom kind handlers
                     on browser side with yaji.userfeedback.registerHandler().
        :param params: Some :samp:`kind`-specific parameters.
        """
        if self.__app.current_request and not self.__app.current_request.runs_downstream:
            raise RuntimeError("User feedback is only available in downstream request handlers.")
        reqid = self.__nextreqid
        self.__nextreqid += 1
        r = {'reqid': reqid, 'kind': kind}
        r.update(params)
        self.__app.appconfig.setconfig("_yj_userfeedback_current", r)
        self.__responses[reqid] = self.__nix
        with self.__responseslock:
            while True:
                answer = self.__responses[reqid]
                if answer != self.__nix:
                    break
                self.__responsescond.wait()
            self.__responses.pop(reqid)
        self.__app.appconfig.setconfig("_yj_userfeedback_current", None)
        return answer

    def _interpret_response(self, request: 'Request') -> None:
        """
        Processes a userfeedback response from browser side.

        :param request: A client request that contains a userfeedback answer.
        """
        reqid = int(request.get_param("reqid"))
        with self.__responseslock:
            if self.__responses[reqid] == self.__nix:
                self.__responses[reqid] = json.loads(request.get_param("answer"))[0]
                self.__responsescond.notify_all()

    def messagedialog(self, message: TrStrOrStrTyping, buttons: Optional[List[TrStrOrStrTyping]] = None) -> int:
        """
        Shows a message dialog to the user and returns the index of the selected button.

        :param message: The message text to show.
        :param buttons: The buttons to offer.
        """
        if buttons is None:
            buttons = [TrStr("OK")]
        return int(self.make_raw_request("message", message=message, buttons=buttons))

    def inputdialog(self, question: TrStrOrStrTyping, defaultanswer: TrStrOrStrTyping = "") -> Optional[str]:
        """
        Shows an input dialog to the user and returns the entered text (or :samp:`None` when cancelled).

        :param question: The question text to show.
        :param defaultanswer: The default answer text that is written to the text field when showing.
        """
        return self.make_raw_request("input", question=question, defaultanswer=defaultanswer)

    def choicedialog(self, question: TrStrOrStrTyping, choices: List[TrStrOrStrTyping]) -> Optional[int]:
        """
        Shows a choice dialog to the user and returns the selected item (or :samp:`None` when cancelled).

        :param question: The question text to show.
        :param choices: The list of choices the user has to select from.
        """
        return self.make_raw_request("choice", question=question, choices=choices)

    def multilineinputdialog(self, question: TrStrOrStrTyping, defaultanswer: TrStrOrStrTyping = "") -> Optional[str]:
        """
        Like :py:meth:`inputdialog` but multi-line capable.

        :param question: The question text to show.
        :param defaultanswer: The default answer text that is written to the text field when showing.
        """
        return self.make_raw_request("multilineinput", question=question, defaultanswer=defaultanswer)

    def filesystemdialog(self, fstype: str = "file", question: TrStrOrStrTyping = "",
                         startpath: str = "/") -> Optional[str]:
        """
        Shows a file/directory selection dialog to the user and returns the path to the selected item (or :samp:`None`).

        :param fstype: The type of filesystem items to select.
        :param question: The question text to show.
        :param startpath: The directory path to start in.
        """
        return self.make_raw_request("filesystem", fstype=fstype, question=question, startpath=startpath)

    def passworddialog(self, question: TrStrOrStrTyping, defaultanswer: TrStrOrStrTyping = "") -> Optional[str]:
        """
        Shows a password dialog to the user and returns the password (or :samp:`None`).

        :param question: The question text to show.
        :param defaultanswer: The default answer text that is written to the password field when showing.
        """
        return self.make_raw_request("password", question=question, defaultanswer=defaultanswer)


if TYPE_CHECKING:  # ugly hack for type annotations
    from .app import *
